class Node:
    pass


class DataField(Node):
    def __init__(self, name, field_name, field_type, setter=None, getter=None):
        self.name = name
        self.field_name = field_name
        self.field_type = field_type
        self.setter = setter
        self.getter = getter


def extra_data_node(root, data: dict):
    """
    root and data should be same level of deep
    Meta attribute should be:
        class Meta:
            # strict, default False if not set
            # fields, list of fields, str or DataField: name - attribute name,
            ##field_name: corresponding key in data node,
            # setter: setter for the attribute, should return value will be set to attr
            # getter: attribute's getter
            strict = False
            fields = [
                "attribute_name",
                DataField(name="attr_2", field_name="field_name", setter=setter, getter=getter)
            ]
    :param root: root object cls
    :param data: the corresponding data node
    :return:
    """
    meta_node = getattr(root, "Meta", None)
    # for serialization should support a non-params constructor
    instance = meta_node.get_instance() if meta_node else root()
    fields = getattr(meta_node, "fields", tuple())
    strict = getattr(meta_node, "strict", False)
    for key, value in data.items():
        for field in fields:
            if isinstance(field, str):
                if strict and key != field:
                    continue
                setattr(instance, field, value)
            elif isinstance(field, DataField) and key == field.field_name:
                setter = field.setter
                _instance = setter(value)
                setattr(instance, field.name, _instance)
    return instance



from functools import wraps


class GeneralException(Exception):
    def __init__(self, code, msg=""):
        self.code = code
        self.msg = msg

    def __eq__(self, other):
        return self.code == other.code


_exceptions = {}


def exception_handle(exceptions: [Exception], handler):
    for ex in _exceptions:
        _exceptions.update({ex: handler})

    def decorator(func):
        @wraps(func)
        def inner(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except tuple(exceptions) as ex:
                return handler(ex)

        return inner

    return decorator


def get_exceptions():
    return list(_exceptions.keys())



# TODO APIReferSchema
# TODO Document by decorator
# TODO
import inspect
from functools import wraps


# OUTCOME: yaml, json reference files


class APIReference:
    def __init__(self):
        self._refer = {}

    def refer(self, *arg, **kwargs):
        path = kwargs.get("path", "")
        methods = kwargs.get("methods", [])
        registered_handlers = []
        prefix = "/"
        if not registered_handlers:
            registered_handlers = []

        def decorator(func):
            if not callable(func):
                raise RuntimeError(f"{func} is not callable")
            module = inspect.getmodule(func).__name__
            self._refer.update(
                {
                    f"{path}": {
                        "handler": f"{module}.{func.__name__}",
                        "methods": methods,
                        "prefix": prefix if prefix else "",
                        "registeredHandlers": registered_handlers
                        if registered_handlers
                        else [],
                    },
                }
            )
            return func

        return decorator



import re
from functools import partial

import genson

Builder = genson.SchemaBuilder


def create_schema(**kwargs):
    return Builder(**kwargs)


class SchemaNode:
    SCHEMA_ATTRIBUTES = (
        "$schema",
        "type",
        "title",
        "required",
        "properties",
        "patternProperties",
        "items",
    )
    SCHEMA_REQUIRED_ATTRIBUTES = (
        "$schema",
        "type",
        "title",
        "required",
        "properties",
    )
    OTHER_ATTRIBUTES = ("patternProperties", "items")

    def __init__(self, **kwargs):
        self._attributes = {}
        self.set_attributes(**kwargs)
        self._register()

    def _set_attr(self, key, value):
        self.set_attributes(**{key: value})

    def _register(self):
        def set_field(key, value):
            self.set_attributes(**{key: value})

        for attr in self.SCHEMA_ATTRIBUTES:
            name = re.sub(r"^[\s,$]+", "", attr)
            name = re.sub(r"[\s,$]+$", "", name)
            self.__dict__.update({f"set_{name}": partial(set_field, name)})

    def validate_attributes(self, **kwargs):
        # required_attrs = set(self.SCHEMA_REQUIRED_ATTRIBUTES) - set(kwargs.keys())
        # if required_attrs:
        #     raise AttributeError(f"Attribute(s) {required_attrs} is(are) required")
        pass

    def set_attributes(self, **kwargs):
        self.validate_attributes(**kwargs)
        for key, value in kwargs.items():
            if key in self.OTHER_ATTRIBUTES and value:
                self._attributes[key] = value
            if key in self.SCHEMA_ATTRIBUTES:
                self._attributes[key] = value

    def to_node(self):
        return self._attributes

    # TODO value validation



if __name__ == "__main__":
    import schema

    builder = schema.Builder()
    schemaobject = {
        "$schema": "http://json-schema.org/schema#",
        "type": "object",
        "properties": {"hi": {"type": ["integer", "string"]}},
    }
    schema_node = schema.SchemaNode(**schemaobject)
    print(schema_node)
